"""
.. module:: errors
    :platform: Linux
    :synopsis: module containing errors raised by kirk.

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""


class KirkException(Exception):
    """
    The most generic exception that is raised by any module. All kirk errors
    derives from this class.
    """


class PluginError(KirkException):
    """
    Raised when plugins operations have failed.
    """


class CommunicationError(KirkException):
    """
    Raised when error occurs during channels communication.
    """


class SUTError(KirkException):
    """
    Raised when error occurs in SUT.
    """


class KernelPanicError(KirkException):
    """
    Raised during kernel panic.
    """


class KernelTaintedError(KirkException):
    """
    Raised when kernel is tainted.
    """


class KernelTimeoutError(KirkException):
    """
    Raised when kernel is not responding anymore.
    """


class FrameworkError(KirkException):
    """
    Raised when an error occurs inside a framework.
    """


class ExporterError(KirkException):
    """
    Raised when an error occurs during Exporter operations.
    """


class LTXError(KirkException):
    """
    Raised when an error occurs during LTX execution.
    """


class SchedulerError(KirkException):
    """
    Raised when an error occurs during Scheduler operations.
    """


class SessionError(KirkException):
    """
    Raised when an error occurs during Session operations.
    """



"""
.. module:: io
    :platform: Linux
    :synopsis: module for handling I/O blocking operations

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

from types import TracebackType
from typing import (
    IO,
    Any,
    AsyncContextManager,
    Optional,
    Type,
    Union,
)

import libkirk


class AsyncFile(AsyncContextManager):
    """
    Handle files in asynchronous way by running operations inside a separate
    thread.
    """

    def __init__(self, filename: str, mode: str = "r") -> None:
        """
        :param filename: Location of the file to open.
        :type filename: str
        :param mode: Mode used to open the file.
        :type mode: str
        """
        self._filename = filename
        self._mode = mode
        self._file = None

        # Optimize: cache mode checks to avoid repeated string operations
        self._is_read_mode = "r" in mode
        self._is_binary_mode = "b" in mode

    async def __aenter__(self) -> Any:
        await self.open()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> bool:
        await self.close()
        return True

    def __aiter__(self) -> Any:
        return self

    async def __anext__(self) -> Optional[Union[str, bytes]]:
        if not self._is_read_mode:
            raise ValueError("File must be open in read mode")

        if not self._file:
            return None

        line = await libkirk.to_thread(self._file.readline)
        if not line:
            raise StopAsyncIteration()

        return line

    async def open(self) -> None:
        """
        Open the file according to the mode.
        """
        if self._file:
            return

        def _open() -> IO[Any]:
            if self._is_binary_mode:
                return open(self._filename, self._mode)
            return open(self._filename, self._mode, encoding="utf-8")

        self._file = await libkirk.to_thread(_open)

    async def close(self) -> None:
        """
        Close the file.
        """
        if not self._file:
            return

        await libkirk.to_thread(self._file.close)
        self._file = None

    async def seek(self, pos: int) -> None:
        """
        Asynchronous version of `seek()`.

        :param pos: Position to search.
        :type pos: int
        """
        if not self._file:
            return

        await libkirk.to_thread(self._file.seek, pos)

    async def tell(self) -> Optional[int]:
        """
        Asynchronous version of `tell()`.

        :return: Current file position or None if file is not open.
        :rtype: int | None
        """
        if not self._file:
            return None

        return await libkirk.to_thread(self._file.tell)

    async def read(self, size: int = -1) -> Optional[Union[str, bytes]]:
        """
        Asynchronous version of `read()`.

        :param size: Amount of data to read. Default is -1, that means all data
            available.
        :type size: int
        :returns: Data that has been read or None if file is not open.
        :rtype: str | bytes | None
        """
        if not self._file:
            return None

        return await libkirk.to_thread(self._file.read, size)

    async def readline(self) -> Optional[Union[str, bytes]]:
        """
        Asynchronous version of `readline()`.

        :returns: Data that has been read or None if file is not open.
        :rtype: str | bytes | None
        """
        if not self._file:
            return None

        return await libkirk.to_thread(self._file.readline)

    async def write(self, data: Union[str, bytes]) -> None:
        """
        Asynchronous version of `write()`.

        :param data: Data to write inside file.
        :type data: str | bytes
        """
        if not self._file:
            return

        await libkirk.to_thread(self._file.write, data)



"""
.. module:: __init__
    :platform: Linux
    :synopsis: application package definition

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

import asyncio
import sys
import typing
from typing import Callable

from libkirk.evt import EventsHandler

# Kirk version
__version__ = "4.1.0"


events = EventsHandler()


def get_event_loop() -> asyncio.AbstractEventLoop:
    """
    Return the current asyncio event loop.
    """
    try:
        return asyncio.get_running_loop()
    except (AttributeError, RuntimeError):
        pass

    try:
        return asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return loop


def create_task(coro: typing.Coroutine) -> asyncio.Task:
    """
    Create a new task.
    """
    return get_event_loop().create_task(coro)


def all_tasks(loop: asyncio.AbstractEventLoop) -> set:
    """
    Return the list of all running tasks for the specific loop.
    """
    # Maintain Python 3.6 compatibility
    if sys.version_info >= (3, 7):
        return asyncio.all_tasks(loop=loop)
    else:
        return asyncio.Task.all_tasks(loop=loop)


def cancel_tasks(loop: asyncio.AbstractEventLoop) -> None:
    """
    Cancel all asyncio running tasks.
    """
    tasks = all_tasks(loop)
    if not tasks:
        return

    tasks_to_cancel = []
    for task in tasks:
        if not (task.cancelled() or task.done()):
            task.cancel()
            tasks_to_cancel.append(task)

    # Maintain Python 3.6 compatibility
    if sys.version_info >= (3, 10):
        # pyrefly: ignore[bad-argument-type]
        loop.run_until_complete(asyncio.gather(*tasks, return_exceptions=True))
    else:
        loop.run_until_complete(
            # pyrefly: ignore[bad-argument-type]
            asyncio.gather(*tasks, loop=loop, return_exceptions=True)
        )

    for task in tasks_to_cancel:
        if not task.cancelled() and task.exception() is not None:
            loop.call_exception_handler(
                {
                    "message": "unhandled exception during asyncio.run() shutdown",
                    "exception": task.exception(),
                    "task": task,
                }
            )


def to_thread(callback: Callable, *args: typing.Any) -> typing.Any:
    """
    Run callback inside a thread. This is useful for blocking I/O operations.
    """
    return get_event_loop().run_in_executor(None, callback, *args)


__all__ = [
    "events",
    "get_event_loop",
]



"""
.. module:: runner
    :platform: Linux
    :synopsis: module containing Runner definition and implementation.

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

import asyncio
import logging
import os
import signal
import sys
import time
from typing import (
    Any,
    Dict,
    List,
    Optional,
)

import libkirk
import libkirk.data
from libkirk.data import (
    Suite,
    Test,
)
from libkirk.errors import (
    KernelPanicError,
    KernelTaintedError,
    KernelTimeoutError,
    KirkException,
    SchedulerError,
)
from libkirk.framework import Framework
from libkirk.results import (
    Results,
    ResultStatus,
    SuiteResults,
    TestResults,
)
from libkirk.sut import (
    SUT,
    IOBuffer,
)


class Scheduler:
    """
    Schedule jobs to run on target.
    """

    @property
    def results(self) -> List[Results]:
        """
        Current results. It's reset before every `schedule` call and
        it's populated when a job completes the execution.

        :return: List of results.
        :rtype: list(Results)
        """
        raise NotImplementedError()

    @property
    def stopped(self) -> bool:
        """
        :return: True when scheduler has been stopped. False otherwise.
        :rtype: bool
        """
        raise NotImplementedError()

    async def stop(self) -> None:
        """
        Stop all running jobs.
        """
        raise NotImplementedError()

    async def schedule(self, jobs: List[Any]) -> None:
        """
        Schedule and execute a list of jobs.

        :param jobs: Object containing jobs definition
        :type jobs: list(object)
        """
        raise NotImplementedError()


class RedirectTestStdout(IOBuffer):
    """
    Redirect test stdout data to UI events and save it.
    """

    def __init__(self, test: Test) -> None:
        self.stdout = ""
        self._test = test

    async def write(self, data: str) -> None:
        await libkirk.events.fire("test_stdout", self._test, data)
        self.stdout += data


class RedirectSUTStdout(IOBuffer):
    """
    Redirect SUT stdout data to UI events.
    """

    def __init__(self, sut: SUT) -> None:
        self._sut = sut

    async def write(self, data: str) -> None:
        await libkirk.events.fire("sut_stdout", self._sut.name, data)


class TestScheduler(Scheduler):
    """
    Schedule and run tests, taking into account status of the kernel
    during their execution, as well as tests timeout.
    """

    STATUS_OK = 0
    TEST_TIMEOUT = 1
    KERNEL_PANIC = 2
    KERNEL_TAINTED = 3
    KERNEL_TIMEOUT = 4

    def __init__(
        self, sut: SUT, framework: Framework, timeout: float = 0.0, max_workers: int = 1
    ) -> None:
        """
        :param sut: Object to communicate with SUT.
        :type sut: SUT
        :param framework: Framework handler.
        :type framework: Framework
        :param timeout: Timeout for tests execution.
        :type timeout: float
        :param max_workers: Maximum number of workers to schedule jobs.
        :type max_workers: int
        """
        if not sut:
            raise ValueError("SUT object is empty")

        if not framework:
            raise ValueError("Framework object is empty")

        self._logger = logging.getLogger("kirk.test_scheduler")
        self._sut = sut
        self._framework = framework
        self._timeout = 0.0 if timeout < 0.0 else timeout
        self._max_workers = 1 if max_workers < 1 else max_workers
        self._results = []
        self._stop_cnt = 0
        self._stopped = False
        self._running_tests_sem = asyncio.Semaphore(1)
        self._schedule_lock = asyncio.Lock()

    async def _get_tainted_status(self) -> tuple:
        """
        Check tainted status of the Kernel.
        """
        code, messages = await self._sut.get_tainted_info()

        for msg in messages:
            if msg:
                self._logger.debug("Kernel tainted (%d): %s", code, msg)
                await libkirk.events.fire("kernel_tainted", msg)

        return code, messages

    async def _write_kmsg(
        self, test: Test, results: Optional[TestResults] = None
    ) -> None:
        """
        If root, we write test information on /dev/kmsg.
        """
        self._logger.info("Writing test information on /dev/kmsg")

        channel = self._sut.get_channel()

        ret = await channel.run_command("id -u")
        if not ret or ret["stdout"] != "0\n":
            self._logger.info("Can't write on /dev/kmsg from user")
            return

        if results:
            message = (
                f"{sys.argv[0]}[{os.getpid()}]: "
                f"{test.name}: end (returncode: {results.return_code})\n"
            )
        else:
            message = (
                f"{sys.argv[0]}[{os.getpid()}]: "
                f"{test.name}: start (command: {test.full_command})\n"
            )

        await channel.run_command(f'echo -n "{message}" > /dev/kmsg')

    @property
    def results(self) -> List[Results]:
        return self._results

    @property
    def stopped(self) -> bool:
        return self._stopped

    async def stop(self) -> None:
        self._logger.info("Stopping tests execution")
        self._stop_cnt += 1

        if self._stop_cnt > 1:
            # by stopping SUT first, we cause scheduler to complete
            # current test immediatelly without waiting
            await self._sut.stop()

        try:
            # we enter in the semaphore queue in order to get highest
            # priority in the tests queue and wait for the running tests
            # to be completed
            async with self._running_tests_sem:
                pass

            self._logger.info("Wait for running tests to be completed")

            async with self._schedule_lock:
                pass
        finally:
            self._stop_cnt = 0
            self._stopped = True

        self._logger.info("All tests have been completed")

    async def _run_test(self, test: Test) -> None:
        """
        Run a single test and populate the results array.
        """
        async with self._running_tests_sem:
            if self._stop_cnt > 0:
                self._logger.info("Test '%s' has been stopped", test.name)
                return

            self._logger.info("Running test %s", test.name)
            self._logger.debug(test)

            await libkirk.events.fire("test_started", test)
            await self._write_kmsg(test, None)

            iobuffer = RedirectTestStdout(test)
            cmd = test.full_command
            start_t = time.time()
            exec_time = 0
            test_data: Dict[str, Any] = {}
            tainted_msg = None
            status = self.STATUS_OK
            channel = self._sut.get_channel()

            try:
                tainted_code1, _ = await self._get_tainted_status()

                # pyrefly: ignore[bad-assignment]
                test_data = await asyncio.wait_for(
                    channel.run_command(
                        cmd, cwd=test.cwd, env=test.env, iobuffer=iobuffer
                    ),
                    timeout=self._timeout,
                )

                if test_data is None:
                    raise SchedulerError("Test command return None")

                tainted_code2, tainted_msg2 = await self._get_tainted_status()
                if tainted_code2 != tainted_code1:
                    self._logger.info("Recognised Kernel tainted: %s", tainted_msg2)

                    tainted_msg = tainted_msg2
                    status = self.KERNEL_TAINTED
            except KernelPanicError:
                exec_time = time.time() - start_t

                self._logger.info("Recognised Kernel panic")
                status = self.KERNEL_PANIC
            except asyncio.TimeoutError:
                exec_time = time.time() - start_t
                status = self.TEST_TIMEOUT

                self._logger.info("Got test timeout. Checking if SUT is still replying")

                try:
                    await asyncio.wait_for(channel.ping(), timeout=10)

                    self._logger.info("SUT replied")
                except asyncio.TimeoutError:
                    status = self.KERNEL_TIMEOUT

            # create test results and save it
            if status not in [self.STATUS_OK, self.KERNEL_TAINTED]:
                test_data = {
                    "name": test.name,
                    "command": test.full_command,
                    "stdout": iobuffer.stdout,
                    "returncode": -1,
                    "exec_time": exec_time,
                }

            # we won't consider tests killed by kirk during forcibly stop,
            # but only if they have been killed by an external application
            # or kernel OOM
            if test_data["returncode"] == -signal.SIGKILL and self._stop_cnt > 1:
                self._logger.info("Test killed: %s", test.name)
                return

            results = await self._framework.read_result(
                test,
                test_data["stdout"],
                test_data["returncode"],
                test_data["exec_time"],
            )

            self._logger.debug("results=%s", results)
            self._results.append(results)

            await libkirk.events.fire("test_completed", results)
            await self._write_kmsg(test, results)

            self._logger.info("Test completed: %s", test.name)
            self._logger.debug(results)

            # raise kernel errors at the end so we can collect test results
            if status == self.KERNEL_TAINTED:
                await libkirk.events.fire("kernel_tainted", tainted_msg)
                raise KernelTaintedError()

            if status == self.KERNEL_PANIC:
                await libkirk.events.fire("kernel_panic")
                raise KernelPanicError()

            if status == self.KERNEL_TIMEOUT:
                await libkirk.events.fire("sut_not_responding")
                raise KernelTimeoutError()

    async def _run_and_wait(self, tests: List[Test]) -> None:
        """
        Run tests one after another.
        """
        if not tests:
            return

        self._logger.info("Scheduling %d tests on single worker", len(tests))

        self._running_tests_sem = asyncio.Semaphore(1)

        for test in tests:
            await self._run_test(test)

    async def _run_parallel(self, tests: List[Test]) -> None:
        """
        Run tests in parallel.
        """
        if not tests:
            return

        self._running_tests_sem = asyncio.Semaphore(self._max_workers)
        coros = [self._run_test(test) for test in tests]

        self._logger.info(
            "Scheduling %d tests on %d workers", len(coros), self._max_workers
        )

        await asyncio.gather(*coros)

    async def schedule(self, jobs: List[Any]) -> None:
        if not jobs:
            raise ValueError("jobs list is empty")

        for job in jobs:
            if not isinstance(job, Test):
                raise ValueError("jobs must be a list of Test")

        async with self._schedule_lock:
            self._logger.info("Check what tests can be run in parallel")

            self._results.clear()

            # Cache filtered lists to avoid redundant list comprehensions
            parallelizable_tests = [test for test in jobs if test.parallelizable]
            sequential_tests = [test for test in jobs if not test.parallelizable]

            try:
                if self._max_workers > 1:
                    await self._run_parallel(parallelizable_tests)
                    await self._run_and_wait(sequential_tests)
                else:
                    await self._run_and_wait(jobs)
            except KirkException as err:
                exc_name = err.__class__.__name__
                self._logger.info("%s caught during tests execution", exc_name)

                async with self._running_tests_sem:
                    pass

                if self._stop_cnt == 0:
                    self._logger.info("Propagating %s exception", exc_name)
                    raise err


class SuiteScheduler(Scheduler):
    """
    The Scheduler class implementation for suites execution.
    This is a special scheduler that schedules suites tests, checking for
    kernel status and rebooting SUT if we have some issues with it
    (i.e. kernel panic).
    """

    def __init__(
        self,
        sut: SUT,
        framework: Framework,
        suite_timeout: float = 0.0,
        exec_timeout: float = 0.0,
        max_workers: int = 1,
    ) -> None:
        """
        :param sut: Object used to communicate with SUT.
        :type sut: SUT
        :param framework: Framework handler.
        :type framework: Framework
        :param suite_timeout: Timeout before stopping testing suite.
        :type suite_timeout: float
        :param exec_timeout: Timeout before stopping single execution.
        :type exec_timeout: float
        :param max_workers: Maximum number of workers to schedule jobs.
        :type max_workers: int
        """
        if not sut:
            raise ValueError("SUT is an empty object")

        if not framework:
            raise ValueError("Framework object is empty")

        suite_timeout = 0.0 if suite_timeout < 0.0 else suite_timeout
        exec_timeout = 0.0 if exec_timeout < 0.0 else exec_timeout
        max_workers = 1 if max_workers < 1 else max_workers

        self._logger = logging.getLogger("kirk.suite_scheduler")
        self._sut = sut
        self._framework = framework
        self._results = []
        self._stop = False
        self._stopped = False
        self._schedule_lock = asyncio.Lock()
        self._reboot_lock = asyncio.Lock()
        self._sut_rebooted = False
        self._suite_timeout = suite_timeout

        self._scheduler = TestScheduler(
            sut=self._sut,
            framework=self._framework,
            timeout=exec_timeout,
            max_workers=max_workers,
        )

    @property
    def results(self) -> List[Results]:
        return self._results

    @property
    def stopped(self) -> bool:
        return self._stopped

    async def stop(self) -> None:
        self._logger.info("Stopping suites execution")

        self._stop = True
        try:
            await self._scheduler.stop()

            async with self._schedule_lock:
                pass
        finally:
            self._stop = False
            self._stopped = True

        self._logger.info("Suites execution has stopped")

    async def _restart_sut(self) -> None:
        """
        Restart the SUT after stopping the tests scheduling.
        """
        async with self._reboot_lock:
            self._logger.info("Rebooting SUT")

            await libkirk.events.fire("sut_restart", self._sut.name)

            iobuffer = RedirectSUTStdout(self._sut)

            await self._scheduler.stop()
            await self._sut.restart(iobuffer=iobuffer)

            self._logger.info("SUT rebooted")

    async def _run_suite(self, suite: Suite) -> None:
        """
        Run a single testing suite and populate the results array.
        """
        self._logger.info("Running suite %s", suite.name)
        self._logger.debug(suite)

        await libkirk.events.fire("suite_started", suite)

        info = await self._sut.get_info()

        start_t = 0.0
        timed_out = False
        exec_times = []
        tests_results: List[TestResults] = []
        tests_left = list(suite.tests)
        reboot_event = asyncio.Event()

        try:
            while not self._stop and tests_left:
                try:
                    start_t = time.time()
                    await asyncio.wait_for(
                        self._scheduler.schedule(tests_left),
                        timeout=self._suite_timeout,
                    )
                except asyncio.TimeoutError:
                    self._logger.info("Testing suite timed out: %s", suite.name)

                    await libkirk.events.fire(
                        "suite_timeout", suite, self._suite_timeout
                    )

                    timed_out = True
                except (KernelPanicError, KernelTaintedError, KernelTimeoutError):
                    if self._reboot_lock.locked():
                        self._logger.info("SUT is rebooting. Waiting...")

                        try:
                            await asyncio.wait_for(reboot_event.wait(), 3600)
                        except asyncio.TimeoutError:
                            self._logger.info("SUT reboot timed out")
                            timed_out = True
                    else:
                        await self._restart_sut()
                        reboot_event.set()
                finally:
                    exec_times.append(time.time() - start_t)
                    # pyrefly: ignore[bad-argument-type]
                    tests_results.extend(self._scheduler.results)

                tests_left.clear()
                completed_test_names = {
                    test_res.test.name for test_res in tests_results
                }
                tests_left.extend(
                    [
                        test
                        for test in suite.tests
                        if test.name not in completed_test_names
                    ]
                )

                if timed_out:
                    for test in tests_left:
                        tests_results.append(
                            TestResults(
                                test=test,
                                failed=0,
                                passed=0,
                                broken=0,
                                skipped=1,
                                warnings=0,
                                exec_time=0.0,
                                retcode=32,
                                stdout="",
                                status=ResultStatus.CONF,
                            )
                        )

                    tests_left.clear()
                    break
        finally:
            suite_exec_time = self._suite_timeout
            if exec_times:
                suite_exec_time = sum(exec_times)

            suite_results = SuiteResults(
                suite=suite,
                tests=tests_results,
                distro=info["distro"],
                distro_ver=info["distro_ver"],
                kernel=info["kernel"],
                cmdline=info["cmdline"],
                arch=info["arch"],
                cpu=info["cpu"],
                swap=info["swap"],
                ram=info["ram"],
            )

            await libkirk.events.fire("suite_completed", suite_results, suite_exec_time)

            self._logger.info("Suite completed")
            self._logger.debug(suite_results)

            self._results.append(suite_results)

    async def schedule(self, jobs: List[Any]) -> None:
        if not jobs:
            raise ValueError("jobs list is empty")

        for job in jobs:
            if not isinstance(job, Suite):
                raise ValueError("jobs must be a list of Suite")

        async with self._schedule_lock:
            self._results.clear()

            for suite in jobs:
                await self._run_suite(suite)



"""
.. module:: main
    :platform: Linux
    :synopsis: main script

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

import argparse
import asyncio
import os
import re
from typing import (
    Dict,
    List,
    Optional,
    Union,
)

import libkirk
import libkirk.com
import libkirk.data
import libkirk.plugin
import libkirk.sut
from libkirk import __version__
from libkirk.com import ComChannel
from libkirk.errors import (
    CommunicationError,
    KirkException,
    SUTError,
)
from libkirk.monitor import JSONFileMonitor
from libkirk.session import Session
from libkirk.sut import SUT
from libkirk.tempfile import TempDir
from libkirk.ui import (
    ParallelUserInterface,
    SimpleUserInterface,
    VerboseUserInterface,
)

# Maximum number of COM instances
MAX_COM_INSTANCES = 128

# return codes of the application
RC_OK = 0
RC_ERROR = 1
RC_INTERRUPT = 130


def _from_params_to_config(params: List[str]) -> Dict[str, str]:
    """
    Return a configuration as dictionary according with input parameters
    given to the commandline option.
    """
    config = {}
    for param in params:
        if "=" not in param:
            raise argparse.ArgumentTypeError(
                f"Missing '=' assignment in '{param}' parameter"
            )

        data = param.split("=", 1)
        key = data[0]
        value = data[1]

        if not key:
            raise argparse.ArgumentTypeError(f"Empty key for '{param}' parameter")

        if not value:
            raise argparse.ArgumentTypeError(f"Empty value for '{param}' parameter")

        config[key] = value

    return config


def _dict_config(
    value: str,
) -> Dict[str, str]:
    """
    Generic dictionary option configuration.
    """
    if value == "help":
        return {"help": ""}

    if not value:
        raise argparse.ArgumentTypeError("Parameters list can't be empty")

    params = value.split(":")

    config = _from_params_to_config(params[1:])
    config["name"] = params[0]

    return config


def _com_config(value: str) -> Optional[Dict[str, str]]:
    """
    Return the list of channels configurations.
    """
    plugins = libkirk.com.get_channels()
    config = _dict_config(value)

    if "help" in config:
        return config

    name = config["name"]

    plugin_names = {p.name for p in plugins}
    if name not in plugin_names:
        raise argparse.ArgumentTypeError(
            f"Can't find communication handler with name '{name}'"
        )

    return config


def _print_plugin_help(
    opt_name: str,
    plugins: Union[List[ComChannel], List[SUT]],
) -> None:
    """
    Print the ``plugins`` help for ``opt_name`` option.
    """
    msg = f"{opt_name} option supports the following syntax:\n"
    msg += "\n\t<name>:<param1>=<value1>:<param2>=<value2>:..\n"
    msg += "\nSupported plugins: | "

    for plugin in plugins:
        msg += f"{plugin.name} | "

    msg += "\n"

    for plugin in plugins:
        if not plugin.config_help:
            msg += f"\n{plugin.name} has not configuration\n"
        else:
            msg += f"\n{plugin.name} configuration:\n"
            for opt, desc in plugin.config_help.items():
                msg += f"\t{opt}: {desc}\n"

    print(msg)


def _iterate_config(value: str) -> int:
    """
    Return the iterate value.
    """
    if not value:
        return 1

    try:
        ret = int(value)
    except TypeError as err:
        raise argparse.ArgumentTypeError("Invalid number") from err

    return max(1, ret)


def _time_config(data: str) -> int:
    """
    Return the time in seconds from '30s', '4m', '5h', '20d' format.
    If no suffix is specified, value is considered in seconds.
    """
    indata = data.strip()

    match = re.search(r"^(?P<value>\d+)\s*(?P<suffix>[smhd]?)$", indata)
    if not match:
        raise argparse.ArgumentTypeError(f"Incorrect time format '{indata}'")

    value = int(match.group("value"))
    suffix = match.group("suffix")

    # Optimize: use dict lookup instead of if/elif chain
    multipliers = {
        "": 1,
        "s": 1,
        "m": 60,
        "h": 3600,
        "d": 86400,  # 3600 * 24
    }

    return value * multipliers.get(suffix, 1)


def _finjection_config(value: str) -> int:
    """
    Return probability of fault injection.
    """
    if not value:
        return 0

    try:
        ret = int(value)
    except TypeError as err:
        raise argparse.ArgumentTypeError("Invalid number") from err

    return max(0, min(100, ret))


def _get_skip_tests(skip_tests: str, skip_file: str) -> str:
    """
    Return the skipped tests regexp.
    """
    skip_parts = []

    if skip_file:
        with open(skip_file, "r", encoding="utf-8") as skip_file_data:
            lines = skip_file_data.readlines()

        toskip = [
            line.rstrip()
            for line in lines
            if line.strip() and not re.search(r"^\s*#", line)
        ]
        if toskip:
            skip_parts.append("|".join(toskip))

    if skip_tests:
        skip_parts.append(skip_tests)

    return "|".join(skip_parts)


def _init_channels(
    args: argparse.Namespace, parser: argparse.ArgumentParser, tmpdir: TempDir
) -> None:
    """
    Initialize channels according to configuration.
    """
    for config in args.com:
        if "id" in config:
            plugin = libkirk.com.clone_channel(config["name"], config["id"])
        else:
            name = config["name"]
            plugin = next(
                (c for c in libkirk.com.get_channels() if c.name == name),
                None,
            )

            assert plugin

        com_config = config.copy()
        com_config["tmpdir"] = tmpdir.abspath

        try:
            # pyrefly: ignore[bad-argument-type]
            plugin.setup(**com_config)
        except CommunicationError as err:
            parser.error(str(err))


def _get_sut(
    args: argparse.Namespace, parser: argparse.ArgumentParser, tmpdir: TempDir
) -> SUT:
    """
    Create and return SUT object.
    """
    sut_config = args.sut.copy()
    sut_config["tmpdir"] = tmpdir.abspath

    sut_name = args.sut["name"]
    sut = next((s for s in libkirk.sut.get_suts() if s.name == sut_name), None)
    if not sut:
        parser.error(f"'{sut_name}' SUT is not available")

    # pyrefly: ignore[missing-attribute]
    sut.optimize = args.optimize_sut

    try:
        # pyrefly: ignore[missing-attribute]
        sut.setup(**sut_config)
    except SUTError as err:
        parser.error(str(err))

    # pyrefly: ignore[bad-return]
    return sut


def _start_session(args: argparse.Namespace, parser: argparse.ArgumentParser) -> None:
    """
    Start the LTP session.
    """
    skip_tests = _get_skip_tests(args.skip_tests, args.skip_file)
    if skip_tests:
        try:
            re.compile(skip_tests)
        except re.error:
            parser.error(f"'{skip_tests}' is not a valid regular expression")

    # check if session can be restored
    restore_dir = args.restore
    if restore_dir and os.path.islink(args.restore):
        restore_dir = os.readlink(args.restore)

    if restore_dir and not os.path.isdir(restore_dir):
        parser.error(f"Can't restore '{args.restore}'. Folder doesn't exist")

    if args.tmp_dir == "":
        tmpdir = TempDir(None)
    elif args.tmp_dir:
        tmpdir = TempDir(args.tmp_dir)
    else:
        tmpdir = TempDir("/tmp")

    # initialize channels
    if args.com:
        _init_channels(args, parser, tmpdir)

    # create SUT
    sut = _get_sut(args, parser, tmpdir)

    # start session
    session = Session(
        tmpdir=tmpdir,
        sut=sut,
        exec_timeout=args.exec_timeout,
        suite_timeout=args.suite_timeout,
        workers=args.workers,
        force_parallel=args.force_parallel,
    )

    # initialize monitor file
    monitor = None
    if args.monitor:
        monitor = JSONFileMonitor(args.monitor)

    # initialize user interface
    if args.workers > 1:
        ParallelUserInterface(args.no_colors)
    elif args.verbose:
        VerboseUserInterface(args.no_colors)
    else:
        SimpleUserInterface(args.no_colors)

    # read tests regex filter
    run_pattern = args.run_pattern
    if run_pattern:
        try:
            re.compile(run_pattern)
        except re.error:
            parser.error(f"'{run_pattern}' is not a valid regular expression")

    async def session_run() -> None:
        """
        Run session then stop events handler.
        """
        try:
            if monitor:
                await monitor.start()

            await session.run(
                command=args.run_command,
                suites=args.run_suite,
                pattern=run_pattern,
                report_path=args.json_report,
                restore_path=restore_dir,
                suite_iterate=args.suite_iterate,
                skip_tests=skip_tests,
                randomize=args.randomize,
                runtime=args.runtime,
                fault_prob=args.fault_injection,
            )
        except asyncio.CancelledError:
            await session.stop()
        finally:
            await libkirk.events.stop()
            if monitor:
                await monitor.stop()

    loop = libkirk.get_event_loop()
    exit_code = RC_OK

    try:
        loop.run_until_complete(
            # pyrefly: ignore[bad-argument-type]
            asyncio.gather(*[libkirk.events.start(), session_run()])
        )
    except KeyboardInterrupt:
        exit_code = RC_INTERRUPT
    except KirkException:
        exit_code = RC_ERROR
    finally:
        try:
            loop.run_until_complete(session.stop())
        except KeyboardInterrupt:
            loop.run_until_complete(session.stop())

        libkirk.cancel_tasks(loop)
        loop.run_until_complete(libkirk.events.stop())

    parser.exit(exit_code)


def run(cmd_args: Optional[List[str]] = None) -> None:
    """
    Entry point of the application.

    :param cmd_args: Command line arguments.
    :type cmd_args: list(str) | None
    """
    currdir = os.path.dirname(os.path.realpath(__file__))

    libkirk.com.discover(os.path.join(currdir, "channels"))
    libkirk.sut.discover(currdir)

    parser = argparse.ArgumentParser(
        description="Kirk - All-in-one Linux Testing Framework"
    )

    generic_opts = parser.add_argument_group("General options")
    generic_opts.add_argument(
        "--version", "-V", action="version", version=f"%(prog)s, {__version__}"
    )
    generic_opts.add_argument(
        "--verbose", "-v", action="store_true", help="Verbose mode"
    )
    generic_opts.add_argument(
        "--no-colors", "-n", action="store_true", help="If defined, no colors are shown"
    )
    generic_opts.add_argument(
        "--tmp-dir", "-d", type=str, default="/tmp", help="Temporary directory"
    )
    generic_opts.add_argument(
        "--restore", "-r", type=str, help="Restore a specific session"
    )
    generic_opts.add_argument(
        "--json-report", "-o", type=str, help="JSON output report"
    )
    generic_opts.add_argument(
        "--monitor", "-m", type=str, help="Location of the monitor file"
    )
    generic_opts.add_argument(
        "--plugins", "-P", type=str, help="Location of custom plugins"
    )

    conf_opts = parser.add_argument_group("Configuration options")
    conf_opts.add_argument(
        "--com",
        "-C",
        type=_com_config,
        action="append",
        help="Communication channel parameters. For help please use '--com help'",
    )
    conf_opts.add_argument(
        "--sut",
        "-u",
        default="default",
        type=lambda x: _dict_config(x),
        help="System Under Test parameters. For help please use '--sut help'",
    )
    conf_opts.add_argument("--skip-tests", "-s", type=str, help="Skip specific tests")
    conf_opts.add_argument(
        "--skip-file",
        "-S",
        type=str,
        help="Skip specific tests using a skip file (newline separated item)",
    )

    exec_opts = parser.add_argument_group("Execution options")
    exec_opts.add_argument("--run-suite", "-f", nargs="*", help="List of suites to run")
    exec_opts.add_argument(
        "--run-pattern", "-p", help="Run all tests matching the regex pattern"
    )
    exec_opts.add_argument("--run-command", "-c", help="Command to run")
    exec_opts.add_argument(
        "--suite-timeout",
        "-T",
        type=_time_config,
        default="1h",
        help="Timeout before stopping the suite (default: 1h)",
    )
    exec_opts.add_argument(
        "--exec-timeout",
        "-t",
        type=_time_config,
        default="1h",
        help="Timeout before stopping a single execution (default: 1h)",
    )
    exec_opts.add_argument(
        "--randomize",
        "-R",
        action="store_true",
        help="Force parallelization execution of all tests",
    )
    exec_opts.add_argument(
        "--runtime",
        "-I",
        type=_time_config,
        default="0",
        help="Set for how long we want to run the session in seconds",
    )
    exec_opts.add_argument(
        "--suite-iterate",
        "-i",
        type=_iterate_config,
        default=1,
        help="Number of times to repeat testing suites",
    )
    exec_opts.add_argument(
        "--workers",
        "-w",
        type=int,
        default=1,
        help="Number of workers to execute tests in parallel",
    )
    exec_opts.add_argument(
        "--force-parallel",
        "-W",
        action="store_true",
        help="Force parallelization execution of all tests",
    )
    exec_opts.add_argument(
        "--fault-injection",
        "-F",
        type=_finjection_config,
        default=0,
        help="Probability of failure (0-100)",
    )
    exec_opts.add_argument(
        "--optimize-sut",
        "-O",
        action="store_true",
        help="Communicate with SUT using commands parallelization (default: false)",
    )

    # output arguments
    # parse comand line
    args = parser.parse_args(cmd_args)

    if args.plugins:
        if not os.path.isdir(args.plugins):
            parser.error(f"'{args.plugins}' plugins directory doesn't exist")

        libkirk.com.discover(args.plugins)
        libkirk.sut.discover(args.plugins)

    if args.com and any("help" in obj for obj in args.com):
        _print_plugin_help("--com", libkirk.com.get_channels())
        parser.exit(RC_OK)

    if args.com and len(args.com) >= MAX_COM_INSTANCES:
        parser.error(f"Maximum number of communication objects is {MAX_COM_INSTANCES}")

    if args.sut and "help" in args.sut:
        _print_plugin_help("--sut", libkirk.sut.get_suts())
        parser.exit(RC_OK)

    if args.json_report and os.path.exists(args.json_report):
        parser.error(f"JSON report file already exists: {args.json_report}")

    if args.run_pattern and not args.run_suite:
        parser.error("--run-pattern must be used with --run-suite")

    if not args.run_suite and not args.run_command:
        parser.error("--run-suite/--run-command are required")

    if args.skip_file and not os.path.isfile(args.skip_file):
        parser.error(f"'{args.skip_file}' skip file doesn't exist")

    if args.tmp_dir and not os.path.isdir(args.tmp_dir):
        parser.error(f"'{args.tmp_dir}' temporary folder doesn't exist")

    _start_session(args, parser)


if __name__ == "__main__":
    run()



"""
.. module:: types
    :platform: Linux
    :synopsis: module that handles types conversion

.. moduleauthor:: Andrea Cervesato <andrea.cervesato@suse.com>
"""

from typing import (
    Any,
    Dict,
    Optional,
    Type,
)


def dict_item(
    data: Dict[str, Any], key: str, cls: Type, default: Optional[Any] = None
) -> Any:
    """
    Extract a value from a dictionary according to the key, ensuring that
    correct type is returned.

    :param data: Dictionary from where we want to extract data.
    :type data: dict
    :param key: Key we are searching for.
    :type key: str
    :param cls: Type we want to extract.
    :type cls: Type
    :param default: Default value.
    :type default: Any | None
    :return: Type of the default value.
    :rtype: Any
    """
    if key not in data:
        return cls(default) if default is not None else None

    val = data[key]

    cls_name = cls.__name__
    numeric_types = {"int", "float"}

    # Check type compatibility (skip for numeric conversions)
    if not isinstance(val, cls) and cls_name not in numeric_types:
        raise TypeError(
            f"dict value must be a {cls_name} but it's {type(val).__name__}"
        )

    return cls(val)
